#ifndef FOURVECTOR_H
#define FOURVECTOR_H

class fourvector {
private:
  double t;
  double x;
  double y;
  double z;
  
public:
  fourvector();
  fourvector(double tv, double xv, double yv, double zv);
  
  inline double x0() const {return t;}
  inline double x1() const {return x;}
  inline double x2() const {return y;}
  inline double x3() const {return z;}

  // length of spatial part
  inline double r() const {return sqrt(r2());}

  // squared length of spatial part
  inline double r2() const {return (x * x + y * y + z * z);}

  // azimuthal angle
  inline double phi() const {
    if (x == 0. && y == 0.){ return 0.; }
    double phi0 = atan2(y, x);
    if (phi0 < 0.){phi0 += 2 * M_PI;}
    return phi0;
  }

  // cosine of azimuthal angle
  inline double cosphi() const {
    if (y != 0. || x != 0.){return x / sqrt(x * x + y * y);}
    else {return 1.;}
  }

  // sine of azimuthal angle
  inline double sinphi() const {
    if (y != 0.){return y / sqrt(x * x + y * y);}
    else {return 0.;}
  }

  // polar angle
  inline double theta() const {return acos(z / r());}

  // cosine of polar angle
  inline double costheta() const {if (r() == 0.){return 1.;} else { return z / r(); }}

  // sine of polar angle
  inline double sintheta() const {return sqrt(1. - pow(z / r(), 2));}

  // squared transverse momentum
  inline double pT2() const {return x * x + y * y;}

  // transverse momentum
  inline double pT() const {return sqrt(pT2());}

  // squared transverse mass (identical with pT2() for massless fourvector)
  inline double ET2() const {if (abs(t) >= abs(z)){return t * t - z * z;} else {return 0.;}}

  // transverse mass (identical with pT() for massless fourvector)
  inline double ET() const {return sqrt(ET2());}

  // m2 should be defined with "return t * t - (x * x + y * y + z * z);" !!!
  // squared invariant mass
  //  inline double m2() const {return t * t - x * x - y * y - z * z;}
  // might lead to slight numerical improvement !!!
  //  inline double m2() const {return t * t - (x * x + y * y + z * z);}
  inline double m2() const {return t * t - r2();}

  // invariant mass
  inline double m() const {return sqrt(m2());}
  //  inline double m() const {return sqrt(t * t - (x * x + y * y + z * z));}

  double eta() const;
  double rapidity() const;
  /*
  inline double eta() const {
    if (z == 0.){return 0.;}
    else if (x == 0. && y == 0.){
      if (z > 0.){return +numeric_limits<double>::infinity();}
      else {return -numeric_limits<double>::infinity();}
    }
    else {
      double eps = (pow(x, 2) + pow(y, 2)) / pow(z, 2);
      if (eps < 1.e-13){
	if (z > 0.){return (0.5 * log((r() + z) / (0.5 * z * eps)));}
	else {return (0.5 * log(-0.5 * z * eps / (r() - z)));}
      }
      else {return = 0.5 * log((r() + z) / (r() - z));}
    }
  }

  inline double rapidity() const {
    if (z == 0.){return 0.;}
    else if (x == 0. && y == 0.){
      if (z > 0){return +numeric_limits<double>::infinity();}
      else {return -numeric_limits<double>::infinity();}
    }
    else {return 0.5 * log((t + z) / (t - z));}
  }
  */

  // scalar product of spatial component with that of a fourvector (fv)
  inline double sp3(fourvector fv) const {return (x * fv.x + y * fv.y + z * fv.z);}
  
  // rotation of a fourvector pointing in z-direction to a fourvector in direction given by (phi / theta)
  inline fourvector rotation_beam(double phi, double costheta) const {
    double sintheta = sqrt(1. - pow(costheta, 2.));
    return fourvector(t, cos(phi) * sintheta * z, sin(phi) * sintheta * z, costheta * z);
  }
  // rotation of an arbitrary fourvector (would rotate a fourvector pointing in z-direction to a fourvector in direction given by (phi / theta) or (frot)
  inline fourvector rotation(double phi, double costheta) const {
    double sintheta = sqrt(1. - pow(costheta, 2.));
    double sinphi = sin(phi);
    double cosphi = cos(phi);
    return fourvector(t, cosphi * costheta * x - sinphi * y + cosphi * sintheta * z, sinphi * costheta * x + cosphi * y + sinphi * sintheta * z, -sintheta * x + costheta * z);
  }
  //  inline fourvector rotation(fourvector frot) const {return rotation(frot.phi(), frot.costheta());}
  inline fourvector rotation(fourvector frot) const {
    double phi, costheta;
    if (frot.r() > 0.){
      costheta = frot.z / frot.r();
      if (frot.x > 0.){phi = atan(frot.y / frot.x);}
      else if (frot.x < 0.){phi = atan(frot.y / frot.x) + M_PI;}
      else {
	if (frot.y > 0.){phi = M_PI / 2.;}
	else if(frot.y < 0.){phi = 3. * M_PI / 2.;}
	else{phi = 0.;}
      }
    }
    else {costheta = 0; phi = 0;}

    //  fourvector t = rotation(phi, costheta);
    //  return t;
    return rotation(phi, costheta);
  }
  // rotation of an arbitrary fourvector (would rotate a fourvector pointing in direction given by (phi / theta) or (frot) to a fourvector in z-direction)
  inline fourvector rotation_inverse(double phi, double costheta) const {
    double sintheta = sqrt(1. - pow(costheta, 2.));
    double sinphi = sin(phi);
    double cosphi = cos(phi);
    return fourvector(t, cosphi * costheta * x + sinphi * costheta * y - sintheta * z, -sinphi * x + cosphi * y, sintheta * cosphi * x + sinphi * sintheta * y + costheta * z);
  }
  //  inline fourvector rotation_inverse(fourvector frot) const {return rotation_inverse(frot.phi(), frot.costheta());}
  inline fourvector rotation_inverse(fourvector frot) const {
    double phi, costheta;
    if (frot.r() > 0.){
      costheta = frot.z / frot.r();
      if (frot.x > 0.){phi = atan(frot.y / frot.x);}
      else if (frot.x < 0.){phi = atan(frot.y / frot.x) + M_PI;}
      else {
	if (frot.y > 0.){phi = M_PI / 2.;}
	else if(frot.y < 0.){phi = 3. * M_PI / 2.;}
	else {phi = 0.;}
      }
    }
    else {costheta = 0; phi = 0;}

    //  fourvector t = rotation_inverse(phi, costheta);
    //  return t;
    return rotation_inverse(phi, costheta);
  }

  // boost of a fourvector in z-direction with boost parameter "beta"
  inline fourvector zboost(double beta) const {
    double gamma = 1. / sqrt(1. - pow(beta, 2.));
    return fourvector(gamma * (t - beta * z), x, y, gamma * (-beta * t + z));
  }

  // boost of a fourvector in x-direction with boost parameter "beta"
  inline fourvector xboost(double beta) const {
    double gamma = 1. / sqrt(1. - pow(beta, 2.));
    return fourvector(gamma * (t - beta * x), gamma * (-beta * t + x), y, z);
  }

  // boost of a fourvector into the rest frame of fourvector "boost" with squared invariant-mass "m2"
  inline fourvector boost(fourvector fv, double m2) const {
    double m = sqrt(m2);
    double sp_iii = sp3(fv);
    return fourvector((t * fv.t - sp_iii) / m, x + fv.x / m * (sp_iii / (m + fv.t) - t), y + fv.y / m * (sp_iii / (m + fv.t) - t), z + fv.z / m * (sp_iii / (m + fv.t) - t));
  }

  // boost of a fourvector into the rest frame of fourvector "boost"
  inline fourvector boost(fourvector fv) const {
    return boost(fv, fv.m2());
  }

  // boost of a fourvector from the frame in which the fourvector "boost" is at rest with squared invariant-mass "m2" 
  // to the frame in which the fourvector "boost"  is provided 
  inline fourvector boostinv(fourvector fv, double m2) const {
    double m = sqrt(m2);
    double gam = fv.t/m;
    double bdotp = - ( x*fv.x + y*fv.y + z*fv.z)/fv.t ;

    return fourvector( gam*(t-bdotp), x-gam*fv.x/fv.t*(gam/(gam+1.)*bdotp-t), y-gam*fv.y/fv.t*(gam/(gam+1.)*bdotp-t), z-gam*fv.z/fv.t*(gam/(gam+1.)*bdotp-t));
  }
  

  // P-transformation
  inline fourvector Pinv() const {return fourvector(t, -x, -y, -z);}

  // P-conserving rotation around the origin of the coordinate system
  inline fourvector Prot() const {return fourvector(t, x, -y, -z);}
  
  // invert time and space components
  inline fourvector operator - () const {return fourvector(-t, -x, -y, -z);}

  // add fourvector (fv)
  inline fourvector operator + (const fourvector & fv) const {return fourvector(t + fv.t, x + fv.x, y + fv.y, z + fv.z);}
  inline fourvector & operator += (const fourvector & fv) {t += fv.t; x += fv.x; y += fv.y; z += fv.z; return *this;}

  // subtract fourvector (fv)
  inline fourvector operator - (const fourvector & fv) const {return fourvector(t - fv.t, x - fv.x, y - fv.y, z - fv.z);}
  inline fourvector & operator -= (const fourvector & fv) {t -= fv.t; x -= fv.x; y -= fv.y; z -= fv.z; return *this;}

  // scalar product with fourvector (fv)
  inline double operator * (const fourvector & fv) const {return t * fv.t - (x * fv.x + y * fv.y + z * fv.z);}

  // check equality with fourvector (fv)
  inline bool operator == (const fourvector & fv) const {
    if ((t == fv.t) && (x == fv.x) && (y == fv.y) && (z == fv.z)){return true;}
    else {return false;}
  }

  // check inequality with fourvector (fv)
  inline bool operator != (const fourvector & fv) const {
    if (*this == fv){return false;}
    else {return true;}
  }

  // multiply fourvector (fv) with double (rn): (rn * fv)
  friend inline fourvector operator * (const double & rn, const fourvector & fv) {return fourvector(fv.t * rn, fv.x * rn, fv.y * rn, fv.z * rn);}
  friend inline fourvector operator * (const fourvector & fv, const double & rn) {return fourvector(fv.t * rn, fv.x * rn, fv.y * rn, fv.z * rn);}

  // divide fourvector (fv) by double (rn)
  friend inline fourvector operator / (const fourvector & fv, const double & rn) {return fourvector(fv.t / rn, fv.x / rn, fv.y / rn, fv.z / rn);}

  // calculate opening angle between fourvectors (fv1) and (fv2)
  friend inline double angle(const fourvector & fv1, const fourvector & fv2) {return acos(fv1.sp3(fv2) / (fv1.r() * fv2.r()));}

  // calculate cosine of opening angle between fourvectors (fv1) and (fv2)
  friend inline double cosangle(const fourvector & fv1, const fourvector & fv2) {return (fv1.x * fv2.x + fv1.y * fv2.y + fv1.z * fv2.z) / (fv1.r() * fv2.r());}

  // calculate sine of opening angle between fourvectors (fv1) and (fv2)
  friend inline double sinangle(const fourvector & fv1, const fourvector & fv2) {return sqrt(1. - pow((fv1.x * fv2.x + fv1.y * fv2.y + fv1.z * fv2.z) / (fv1.r() * fv2.r()), 2));}
  
  /*
  friend inline double R(const fourvector & fv1, const fourvector & fv2) {}
  friend inline double R2_eta(const fourvector & fv1, const fourvector & fv2) {}
  friend inline double R2_rapidity(const fourvector & fv1, const fourvector & fv2) {}
  friend inline double R2_cosheta_cosphi(const fourvector & fv1, const fourvector & fv2) {}
  friend inline double R2_coshrapidity_cosphi(const fourvector & fv1, const fourvector & fv2) {}
  */
  friend double R(const fourvector & fv1, const fourvector & fv2);
  friend double R2_eta(const fourvector & fv1, const fourvector & fv2);
  friend double R2_rapidity(const fourvector & fv1, const fourvector & fv2);
  friend double R2_cosheta_cosphi(const fourvector & fv1, const fourvector & fv2);
  friend double R2_coshrapidity_cosphi(const fourvector & fv1, const fourvector & fv2);

  // opening angle with fourvector (fv)
  inline double angle(const fourvector & fv) const {return acos(sp3(fv) / (r() * fv.r()));}
  
  
  /*
  inline bool isnan(){ if (munich_isnan(t) || munich_isnan(x) || munich_isnan(y) || munich_isnan(z)){ return true; } else { return false; } }
  inline bool isinf(){ if (munich_isinf(t) || munich_isinf(x) || munich_isinf(y) || munich_isinf(z)){ return true; } else { return false; } }
  inline bool isinfnan(){ if (this->isnan() || this->isinf()){ return true; } else { return false; } }
  */

  bool isnan();
  bool isinf();
  bool isinfnan();

  friend std::ostream & operator << (std::ostream &, const fourvector & fv);
};

#endif
