#include "header.hpp"

fourvector::fourvector(){
  t = 0;
  x = 0;
  y = 0;
  z = 0;
}
fourvector::fourvector(double r0, double r1, double r2, double r3){
  t = r0;
  x = r1;
  y = r2;
  z = r3;
}

// destructor needed ???
//fourvector::~fourvector();


double fourvector::eta() const {
  static Logger logger("fourvector::eta");
  logger.called();

  // old implementation: //  double eta = -log(tan((theta() / 2.)));
  double _eta = 0.;
  if (z != 0.){
    double eps = (pow(x, 2) + pow(y, 2)) / pow(z, 2);
    if (eps == 0.){
      if (z > 0){return +numeric_limits<double>::infinity();}
      else {return -numeric_limits<double>::infinity();}
    }
    if (eps < 1.e-13){
      if (z > 0.){_eta = 0.5 * log((r() + z) / (0.5 * z * eps));}
      else {_eta = 0.5 * log(-0.5 * z * eps / (r() - z));}
    }
    else {_eta = 0.5 * log((r() + z) / (r() - z));}
    if (munich_isnan(_eta) || abs(_eta) > 1.e99) {
      logger << LOG_WARN << "computation unstable; eta = " << _eta << ", eps = " << eps << endline;
      logger << LOG_WARN << (*this) << endline;
      logger << LOG_WARN << "r = " << r() << ", z = " << z << ", old implementation: eta = " << -log(tan((theta() / 2.))) << endline_flush;
    }
  }

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "return " << _eta << endline_flush; }
  logger.finished();
  return _eta;
}


double fourvector::rapidity() const {
  static Logger logger("fourvector::rapidity");
  logger.called();
  
  double qq = z / t;
  
  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "return " << .5 * log((qq + 1) / (1 - qq)) << endline_flush; }
  logger.finished();
  return .5 * log((1. + qq) / (1. - qq));
}


// calculate distance R (defined via pseudo-rapidity) of fourvectors "fv1" and "fv2"
// replace by functions that explicitly refer to y or eta !!!
double R(const fourvector & fv1, const fourvector & fv2){
  static Logger logger("fourvector::R");
  logger.called();
  
  double delta_phi, R;
  delta_phi = abs(fv1.phi() - fv2.phi());
  if (delta_phi > pi){delta_phi = 2. * pi - delta_phi;}
  R = sqrt(pow(fv1.eta() - fv2.eta(), 2.) + pow(delta_phi, 2.));

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "return " << R << endline_flush; }
  logger.finished();
  return R;
}


double R2_eta(const fourvector & fv1, const fourvector & fv2){
  static Logger logger("fourvector::R2_eta");
  logger.called();
  
  double R2_eta;
  double delta_phi;
  delta_phi = abs(fv1.phi() - fv2.phi());
  if (delta_phi > pi){delta_phi = 2 * pi - delta_phi;}
  R2_eta = pow(fv1.eta() - fv2.eta(), 2) + pow(delta_phi, 2);
  // TODO: could be optimized further

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "return " << R2_eta << endline_flush; }
  logger.finished();
  return R2_eta;
}


double R2_rapidity(const fourvector & fv1, const fourvector & fv2){
  static Logger logger("fourvector::R2_rapidity");
  logger.called();
  
  double R2_rapidity;
  double delta_phi;
  delta_phi = abs(fv1.phi() - fv2.phi());
  if (delta_phi > pi){delta_phi = 2 * pi - delta_phi;}
  R2_rapidity = pow(fv1.rapidity() - fv2.rapidity(), 2) + pow(delta_phi, 2);

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "return " << R2_rapidity << endline_flush; }
  logger.finished();
  return R2_rapidity;
}


double R2_cosheta_cosphi(const fourvector & fv1, const fourvector & fv2){
  static Logger logger("fourvector::R2_cosheta_cosphi");
  logger.called();
  
  double R2_cosheta_cosphi;
  double delta_phi;
  delta_phi = abs(fv1.phi() - fv2.phi());
  if (delta_phi > pi){delta_phi = 2 * pi - delta_phi;}
  R2_cosheta_cosphi = 2 * (cosh(fv1.eta() - fv2.eta()) - cos(delta_phi));

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "return " << R2_cosheta_cosphi << endline_flush; }
  logger.finished();
  return R2_cosheta_cosphi;
}


double R2_coshrapidity_cosphi(const fourvector & fv1, const fourvector & fv2){
  static Logger logger("fourvector::R2_coshrapidity_cosphi");
  logger.called();
  
  double R2_coshrapidity_cosphi;
  double delta_phi;
  delta_phi = abs(fv1.phi() - fv2.phi());
  if (delta_phi > pi){delta_phi = 2 * pi - delta_phi;}
  R2_coshrapidity_cosphi = 2 * (cosh(fv1.rapidity() - fv2.rapidity()) - cos(delta_phi));

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "return " << R2_coshrapidity_cosphi << endline_flush; }
  logger.finished();
  return R2_coshrapidity_cosphi;
}


bool fourvector::isnan(){
  if (munich_isnan(t) || munich_isnan(x) || munich_isnan(y) || munich_isnan(z)){return true;}
  else {return false;}
}


bool fourvector::isinf(){
  if (munich_isinf(t) || munich_isinf(x) || munich_isinf(y) || munich_isinf(z)){return true;}
  else {return false;}
}


bool fourvector::isinfnan(){
  if (this->isnan() || this->isinf()){return true;}
  else {return false;}
}


ostream & operator << (ostream & s, const fourvector & fv){
  s << "( " << setw(21) << setprecision(15) << fv.t << " ; " << setw(21) << setprecision(15) << fv.x << " , " << setw(21) << setprecision(15) << fv.y << " , " << setw(21) << setprecision(15) << fv.z <<" )";
  //  s << "( " << setw(14) << setprecision(8) << fv.t << " ; " << setw(14) << setprecision(8) << fv.x << " , " << setw(14) << setprecision(8) << fv.y << " , " << setw(14) << setprecision(8) << fv.z <<" )";
  return s;
}
