#include "header.hpp"
#include "gsl/gsl_sf_bessel.h"
#include "gsl/gsl_integration.h"

void observable_set::calculate_coefficients_Itilde(const double xmio, const int order) {
  static Logger logger("observable_set::calculate_coefficients_Itilde");
  logger.called();

  double argum = b0_Eulergamma * xmio;
  double logx = log(xmio);

  double zbk0 = zbesselk0(argum);
  double zbk1 = zbesselk1(argum);
  double zbk2 = zbesselk2(argum);
  double zbk3 = zbesselk3(argum);

  LL[1] = -zbk0 / pow(xmio, 2);
  LL[2] = 2. / pow(xmio, 2) * (zbk0 * logx - zbk1);
  if (order > 1){
    double log2x = pow(logx, 2);
    LL[3] = -3. / pow(xmio, 2) *(zbk0 * (log2x - zeta2) - 2. * zbk1 * logx + zbk2);
    LL[4] = 4. / pow(xmio, 2) *(zbk0 * (pow(logx, 3) - 3. * zeta2 * logx + 2 * zeta3) - 3. * zbk1 * (log2x - zeta2) + 3. * zbk2 * logx - zbk3);
  }

  logger.finished();
}


void observable_set::calculate_integrated_Itilde(const int order){
  static Logger logger("observable_set::calculate_integrated_Itilde");
  logger.called();

  gsl_integration_workspace * w = gsl_integration_workspace_alloc (10000);

  double result;
  double error;

  gsl_function F;
  F.params = NULL;

  double xmin = min_qTcut / 100.;
  double xstep = step_qTcut / 100.;
  double xmax = 600.;  //  default starting value ~ inifinity

  // needs to be modified for logarithmic rcut binning (use already available value_qTcut ???) !!!
  for (int i_q = n_qTcut - 1; i_q >= 0; i_q--){
    double xcut = xmin + i_q * xstep;
    if (i_q < n_qTcut - 1){xmax = xmin + (i_q + 1) * xstep;}
      
    F.function = &Itilde1;
    gsl_integration_qags (&F, xcut, xmax, 0, 1.e-7, 10000, w, &result, &error);
    int_I[1][i_q] = result;
    if (i_q < n_qTcut - 1){
      int_I[1][i_q] += int_I[1][i_q + 1];
      error_I[1][i_q] = sqrt(error_I[1][i_q + 1] * error_I[1][i_q + 1] + error * error);
    }
    else {error_I[1][i_q] = error;}

    F.function = &Itilde2;
    gsl_integration_qags (&F, xcut, xmax, 0, 1.e-7, 10000, w, &result, &error);
    int_I[2][i_q] = result;
    if (i_q < n_qTcut - 1){
      int_I[2][i_q] += int_I[2][i_q + 1];
      error_I[2][i_q] = sqrt(error_I[2][i_q + 1] * error_I[2][i_q + 1] + error * error);
    }
    else {error_I[2][i_q] = error;}

    if (order > 1){
      F.function = &Itilde3;
      gsl_integration_qags (&F, xcut, xmax, 0, 1.e-7, 10000, w, &result, &error);
      int_I[3][i_q] = result;
      if (i_q < n_qTcut - 1){
	int_I[3][i_q] += int_I[3][i_q + 1];
	error_I[3][i_q] = sqrt(error_I[3][i_q + 1] * error_I[3][i_q + 1] + error * error);
      }
      else {error_I[3][i_q] = error;}

      F.function = &Itilde4;
      gsl_integration_qag (&F, xcut, xmax, 0, 1.e-8, 10000, 6, w, &result, &error);
      int_I[4][i_q] = result;
      if (i_q < n_qTcut - 1){
	int_I[4][i_q] += int_I[4][i_q + 1];
	error_I[4][i_q] = sqrt(error_I[4][i_q + 1] * error_I[4][i_q + 1] + error * error);
      }
      else {error_I[4][i_q] = error;}
    }
  }

  stringstream temp_ss;
  for (int i_q = 0; i_q < n_qTcut; i_q++){
    double xcut = xmin + i_q * xstep;
    temp_ss << "rcut = " << setw(12) << setprecision(5) << xcut;
    temp_ss << "   I1: " << setprecision(15) << setw(23) << int_I[1][i_q] << " +- " << setprecision(15) << setw(23) << error_I[1][i_q];
    temp_ss << "   I2: " << setprecision(15) << setw(23) << int_I[2][i_q] << " +- " << setprecision(15) << setw(23) << error_I[2][i_q];
    if (order > 1){
      temp_ss << "   I3: " << setprecision(15) << setw(23) << int_I[3][i_q] << " +- " << setprecision(15) << setw(23) << error_I[3][i_q];
      temp_ss << "   I4: " << setprecision(15) << setw(23) << int_I[4][i_q] << " +- " << setprecision(15) << setw(23) << error_I[4][i_q];
    }
    temp_ss << endline;
  }
  logger << LOG_DEBUG << "Precomputing Itilde integrals" << endline << temp_ss.str() << endline_flush;


  logger.finished();
}
