#include "header.hpp"

#include <stdlib.h>
#include <stdio.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_multifit_nlinear.h>

struct data {
  size_t n;
  double * t;
  double * y;
};


int function_quadratic_f (const gsl_vector * x, void *data, gsl_vector * f){
  size_t n = ((struct data *)data)->n;
  double *t = ((struct data *)data)->t;
  double *y = ((struct data *)data)->y;

  vector<double> Xj(3);
  for (size_t i = 0; i < Xj.size(); i++){ Xj[i] = gsl_vector_get (x, i); }
  for (size_t i_n = 0; i_n < n; i_n++){
    double Yi = Xj[2] * pow(t[i_n], 2) + Xj[1] * t[i_n] + Xj[0];
    gsl_vector_set (f, i_n, Yi - y[i_n]);
  }

  return GSL_SUCCESS;
}


int function_quadratic_df (const gsl_vector * x, void *data, gsl_matrix * J){
  size_t n = ((struct data *)data)->n;
  double *t = ((struct data *)data)->t;

  for (size_t i_n = 0; i_n < n; i_n++){
    // Jacobian matrix J(i,j) = dfi / dxj,
    // where fi = (Yi - yi) / sigma[i],
    //       Yi = Xj[2] * pow(t[i], 2) + Xj[1] * t[i] + Xj[0]
    // and the xj are the parameters Xj[]
    gsl_matrix_set (J, i_n, 2, pow(t[i_n], 2));
    gsl_matrix_set (J, i_n, 1, t[i_n]);
    gsl_matrix_set (J, i_n, 0, 1.);
  }

  return GSL_SUCCESS;
}


int function_quadratic_log1_f (const gsl_vector * x, void *data, gsl_vector * f){
  size_t n = ((struct data *)data)->n;
  double *t = ((struct data *)data)->t;
  double *y = ((struct data *)data)->y;
  vector<double> Xj(5);
  for (size_t i = 0; i < Xj.size(); i++){ Xj[i] = gsl_vector_get (x, i); }
  for (size_t i_n = 0; i_n < n; i_n++){
    double Yi = Xj[2] * pow(t[i_n], 2) + Xj[4] * log(t[i_n]) * pow(t[i_n], 2) + Xj[1] * t[i_n] + Xj[3] * log(t[i_n]) * t[i_n] + Xj[0];
    gsl_vector_set (f, i_n, Yi - y[i_n]);
  }

  return GSL_SUCCESS;
}


int function_quadratic_log1_df (const gsl_vector * x, void *data, gsl_matrix * J){
  size_t n = ((struct data *)data)->n;
  double *t = ((struct data *)data)->t;

  for (size_t i_n = 0; i_n < n; i_n++) {
    // Jacobian matrix J(i,j) = dfi / dxj,
    // where fi = (Yi - yi) / sigma[i],
    //       Yi = Xj[2] * pow(t[i_n], 2) + Xj[4] * log(t[i_n]) * pow(t[i_n], 2) + Xj[1] * t[i_n] + Xj[3] * log(t[i_n]) * t[i_n] + Xj[0] 
    // and the xj are the parameters Xj[]
    gsl_matrix_set (J, i_n, 0, 1.);
    gsl_matrix_set (J, i_n, 1, t[i_n]);
    gsl_matrix_set (J, i_n, 2, pow(t[i_n], 2));
    gsl_matrix_set (J, i_n, 3, log(t[i_n]) * t[i_n]);
    gsl_matrix_set (J, i_n, 4, log(t[i_n]) * pow(t[i_n], 2));
  }

  return GSL_SUCCESS;
}


void callback(const size_t iter, void *params, const gsl_multifit_nlinear_workspace *w){
  //  gsl_vector *f = gsl_multifit_nlinear_residual(w);
  //  gsl_vector *x = gsl_multifit_nlinear_position(w);
  double rcond;

  /* compute reciprocal condition number of J(x) */
  gsl_multifit_nlinear_rcond(&rcond, w);

  /*
  fprintf(stderr, "iter %2zu: A = %.4f, lambda = %.4f, b = %.4f, cond(J) = %8.4f, |f(x)| = %.4f\n",
          iter,
          gsl_vector_get(x, 0),
          gsl_vector_get(x, 1),
          gsl_vector_get(x, 2),
          1.0 / rcond,
          gsl_blas_dnrm2(f));
  */
}


int summary_list::extrapolation_function_gsl(vector<double> & xvalue, double & min_extrapolation, double & max_extrapolation, vector<double> & data_result, vector<double> & data_deviation2, vector<double> & X, vector<double> & dX, double & chi2_dof, int order_fit){
  static Logger logger("summary_list::extrapolation_function_gsl");
  logger.called();

  const gsl_multifit_nlinear_type *T = gsl_multifit_nlinear_trust;
  gsl_multifit_nlinear_workspace *w;
  gsl_multifit_nlinear_fdf fdf;
  gsl_multifit_nlinear_parameters fdf_params = gsl_multifit_nlinear_default_parameters();

  size_t n_value = 0;
  int x_q_min = -1;
  for (int i_q = 0; i_q < xvalue.size(); i_q++){
    if (xvalue[i_q] < min_extrapolation){ continue; }
    if (x_q_min == -1){ x_q_min = i_q; }
    if (xvalue[i_q] > max_extrapolation){ break; }
    n_value++;
  }
  if (x_q_min == -1){ logger << LOG_FATAL << "No reasonable rcut range chosen." << endline_flush; exit(1); }

  size_t n_parameter = 0;
  if (order_fit == 2){
    n_parameter = 3;
    fdf.f = function_quadratic_f;
    fdf.df = function_quadratic_df;   /* set to NULL for finite-difference Jacobian */
  }
  else if (order_fit == 3){
    // r*log(r) and r²log(r) terms on top of quadratic function
    n_parameter = 5;
    fdf.f = function_quadratic_log1_f;
    fdf.df = function_quadratic_log1_df;   /* set to NULL for finite-difference Jacobian */
  }
  
  double * x_init;
  x_init = new double[n_parameter];
  for (int i = 0; i < n_parameter; i++){ x_init[i] = 1.; }

  gsl_vector *f;
  gsl_matrix *J;
  gsl_matrix *covar = gsl_matrix_alloc (n_parameter, n_parameter);
  double t[n_value], y[n_value], weights[n_value];
  struct data d = { n_value, t, y };
  
  gsl_vector_view x = gsl_vector_view_array (x_init, n_parameter);
  gsl_vector_view wts = gsl_vector_view_array(weights, n_value);
  double chisq;
  double chisq0;
  int status;
  int info;

  const double xtol = 1.e-8;
  const double gtol = 1.e-8;
  const double ftol = 0.;

  gsl_rng_env_setup();

  fdf.fvv = NULL;     /* not using geodesic acceleration */
  fdf.n = n_value;
  fdf.p = n_parameter;
  fdf.params = &d;

  /* this is the data to be fitted */
  for (int i_n = 0; i_n < n_value; i_n++){
    int i_q = x_q_min + i_n;
    if (data_deviation2[i_q] != 0.){
      t[i_n] = xvalue[i_q];
      weights[i_n] = 1. / data_deviation2[i_q];
      y[i_n] = data_result[i_q];
      //      printf ("data: %g %g %g\n", t[i_n], y[i_n], weights[i_n]);
    }
  }
  
  /* allocate workspace with default parameters */
  w = gsl_multifit_nlinear_alloc (T, &fdf_params, n_value, n_parameter);

  /* initialize solver with starting point and weights */
  gsl_multifit_nlinear_winit (&x.vector, &wts.vector, &fdf, w);

  /* compute initial cost function */
  f = gsl_multifit_nlinear_residual(w);
  gsl_blas_ddot(f, f, &chisq0);

  /* solve the system with a maximum of 100 iterations */
  status = gsl_multifit_nlinear_driver(100, xtol, gtol, ftol, callback, NULL, &info, w);

  /* compute covariance of best fit parameters */
  J = gsl_multifit_nlinear_jac(w);
  gsl_multifit_nlinear_covar (J, 0., covar);

  /* compute final cost */
  gsl_blas_ddot(f, f, &chisq);

  X.resize(n_parameter, 0.);
  dX.resize(n_parameter, 0.);
  for (size_t i = 0; i < X.size(); i++){
    X[i] = gsl_vector_get(w->x, i);
    dX[i] = sqrt(gsl_matrix_get(covar, i, i));
  }

  //#define FIT(i) gsl_vector_get(w->x, i)
  //#define ERR(i) sqrt(gsl_matrix_get(covar,i,i))

  double dof = n_value - n_parameter;
  chi2_dof = chisq / dof;
  double c = GSL_MAX_DBL(1, sqrt(chi2_dof));
  logger << LOG_INFO << "range [" << setw(5) << min_extrapolation << " : " << setw(5) << max_extrapolation << "] :   " << "b = " << setw(15) << setprecision(8) << X[0] << " +- " << setw(15) << setprecision(8) << dX[0] << " (" << setw(15) << setprecision(8) << c * dX[0] << ")" << "   " << "chisq_dof = " << setw(15) << setprecision(8) << chi2_dof << "   " << "status = " << status << endline_flush;

  //  fprintf (stderr, "status = %s\n", gsl_strerror (status));

  gsl_multifit_nlinear_free (w);
  gsl_matrix_free (covar);
  //  gsl_rng_free (r);

  logger.finished();
  return 0;
}
