#pragma once

#include <iostream>
#include <map>

namespace FivePointAmplitudes {

template <typename T> struct sij5 {
    static constexpr size_t N = 5;
    std::map<std::pair<int, int>, T> _s;

    sij5(const std::array<T, N>&);

    /**
     * Get s[i,j], 1-based
     */
    T get(int i, int j) const {
        if (j < i) std::swap(i, j);
        return _s.at({i, j});
    }
    T operator()(int i, int j) const { return get(i, j); }

    /**
     * Return a copy of *this with permuted momenta according to
     * @param permutation
     */
    sij5<T> permute(const std::array<int, N>&);

    operator std::array<T,N>() const;
};

template <typename T> sij5<T>::sij5(const std::array<T, N>& v) {
    for (size_t i = 1; i <= N; ++i) { _s[{i, i}] = T(0); }
    _s[{1, 2}] = v[0];
    _s[{2, 3}] = v[1];
    _s[{3, 4}] = v[2];
    _s[{4, 5}] = v[3];
    _s[{1, 5}] = v[4];

    _s[{1, 3}] = -_s[{1, 2}] - _s[{2, 3}] + _s[{4, 5}];
    _s[{1, 4}] = -_s[{1, 5}] + _s[{2, 3}] - _s[{4, 5}];
    _s[{2, 4}] = _s[{1, 5}] - _s[{2, 3}] - _s[{3, 4}];
    _s[{2, 5}] = -_s[{1, 2}] - _s[{1, 5}] + _s[{3, 4}];
    _s[{3, 5}] = _s[{1, 2}] - _s[{3, 4}] - _s[{4, 5}];
}

template <typename T> sij5<T> sij5<T>::permute(const std::array<int, N>& perm) {

    std::array<T, N> new_v;

    for (size_t i = 0; i < N; ++i) {
        auto j = (i + 1) % N;
        new_v[i] = get(perm[i], perm[j]);
    }

    return sij5<T>(new_v);
}

template <typename T> sij5<T>::operator std::array<T, N>() const {
    std::array<T, N> toret;
    for (size_t i = 0; i < N; ++i) {
        auto j = (i + 1) % N;
        toret[i] = get(i + 1, j + 1);
    }
    return toret;
}

template <typename T> std::ostream& operator<<(std::ostream& s, const sij5<T>& sij) {
    s << "sijs : \n";
    for (auto& it : sij._s) { s << "s[" << it.first.first << "," << it.first.second << "] = " << it.second << "\n"; }
    return s;
}

} // namespace FivePointAmplitudes
