#pragma once

#include "FunctionsManager.h"
#include "Kinematics.h"

namespace FivePointAmplitudes {

namespace detail {
    // Neumaier summation (for improved numerical stability) [doi:10.1002/zamm.19740540106]
template <typename Tcontainter> auto neumaier_sum(const Tcontainter& values) {
    using T = typename Tcontainter::value_type;

    switch (values.size()) {
        case 0: return T(0);
        case 1: return values[0];
        case 2: return values[0] + values[1];
        default: break;
    }

    T sum{0};
    T c{0};

    for (const auto& si : values) {
        using std::abs;
        using std::get;
        auto t = sum + si;
        if (abs(sum) >= abs(si)) { c += ((sum - t) + si); }
        else {
            c += ((si - t) + sum);
        }
        sum = t;
    }
    sum += c;
    return sum;
}

template <typename T> T stof(const char* i) { return T(i);}
template <> inline double stof(const char* i) {return std::atof(i);}

} // detail

template <size_t NLETTERS, typename T, typename Tm, typename Tc_basis, size_t NCOEFFS, typename Tf_basis, size_t NFUNCS>
std::complex<T> eval_remainder(std::array<T, 5> v_in, int str5, T musq, const std::array<Tc_basis, NCOEFFS>& coeff_basis, const std::array<Tf_basis,NFUNCS>& functions_basis,
                               const Tm& coefficients, std::shared_ptr<FunctionsManager> f_manager, T tr5norm) {
    using TC = std::complex<T>;
    using std::array;

    // The coefficients of all functions in all remainders are by construction dimensionless,
    // so for better numerical stability we rescale them.
    // We could choose the scale ourselves here, but we can also assume that the renormalization
    // scale has been chosen naturally, so we use it instead.
    for (auto& v : v_in) { v /= musq; }
    tr5norm /= musq;
    tr5norm *= tr5norm;

    Kin<T> kin(v_in);

    array<T, NLETTERS> iW;
    for (size_t i = 0; i < NLETTERS; ++i) { iW[i] = 1 / (kin.W[i].real()); }
    {
        T d5 = kin.SqrtDelta.imag();
        d5 *= d5;
        iW[NLETTERS - 1] = (-1) / d5;
    }

    array<T, NCOEFFS> coeffs;

    for (size_t i = 0; i < NCOEFFS; ++i) {
        coeffs[i] = coeff_basis[i](v_in, iW);
    }

    const auto& function_values = f_manager->eval(kin);
    array<TC, NFUNCS> functions_basis_values;

    for (size_t i = 0; i < NFUNCS; ++i) {
        TC c{1};
        for (const auto& el : functions_basis[i]) {
            switch (el.first) {
                case f_type::f: c *= function_values.at(el.second); break;
                case f_type::bc: c *= bc<T>.at(el.second); break;
                case f_type::bci: c *= TC(0, bc<T>.at(el.second)); break;
                case f_type::str5: c *= str5; break;
                case f_type::one: break;
            }
        }
        functions_basis_values[i] = std::move(c);
    }

    auto sum_coeff = [&coeffs](const auto& list) {
        using std::get;
        std::vector<T> re_values;
        re_values.reserve(list.size());
        for (const auto& it : list) { re_values.push_back((coeffs[get<int>(it)] * get<T>(it))); }
        return detail::neumaier_sum(re_values);
    };

    // Neumaier summation (for improved numerical stability) [doi:10.1002/zamm.19740540106]
    TC sum{0};
    TC c{0};

    for (size_t i = 0; i < NFUNCS; ++i){
        using std::abs;

        T re = sum_coeff(coefficients[i].first);
        T im = sum_coeff(coefficients[i].second);

        im *= str5;
        im *= kin.SqrtDelta.imag();
        im /= tr5norm;
        TC si{re, im};

        si *= functions_basis_values[i];

        auto t = sum + si;
        if (abs(sum) >= abs(si)) { c += ((sum - t) + si); }
        else {
            c += ((si - t) + sum);
        }
        sum = t;
    }
    sum += c;
    return sum;
}

template <size_t NLETTERS, typename T, typename Tm, typename Tc_basis, size_t NCOEFFS, typename Tf_basis, size_t NFUNCS>
std::complex<T> eval_remainder(std::array<T, 5> v_in, int str5, T musq, const std::array<Tc_basis, NCOEFFS>& coeff_basis,
                               const std::array<Tf_basis, NFUNCS>& functions_basis, const Tm& coefficients, std::shared_ptr<FunctionsManager> f_manager) {
    return eval_remainder(v_in, str5, musq, coeff_basis, functions_basis, coefficients, f_manager, musq);
}

} // namespace FivePointAmplitudes

